# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2007 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

from PrefixScanner import PrefixScanner
from Token import Token
from proposals import VariantProposal, ChoiceProposal
from CommandStore import Command, Variant, OptionalArgument, MandatoryArgument, VectorArgument
from LaTeXPlugin.Logger import debug, error

class PrefixParser(object):
	"""
	This parses a prefix like "\begin{" or just "\a" and generates
	matching proposals, an argument hint and a doc hint
	"""
	
	S_IDLE, S_COMMAND_START, S_COMMAND, S_MANDATORY_ARGUMENT_START, S_OPTIONAL_ARGUMENT_START, \
			S_VECTOR_ARGUMENT_START, S_MANDATORY_ARGUMENT, S_OPTIONAL_ARGUMENT, \
			S_VECTOR_ARGUMENT = range(9)
	
	def __init__(self, command_store):
		self.__command_store = command_store
		self.__scanner = PrefixScanner()
		
		self.proposals = []
		self.hint = None
		
	def parse(self, prefix):
		self.subprefix = prefix
		self.__scanner.set_prefix(prefix)
		
		self.proposals = []
		self.hint = None
		
		state = self.S_IDLE
		
		for token in self.__scanner:
			
			if state == self.S_IDLE:
				if token.type == Token.T_BACKSLASH:
					state = self.S_COMMAND_START
			
			elif state == self.S_COMMAND_START:
				if token.type == Token.T_NAME:
					self.subprefix = token.value
					variant_signature = []				# argument signature
					state = self.S_COMMAND
				elif token.type == Token.T_END:
					# \
					#  ^
					for command in self.__command_store.itervalues():
						for variant in command.variants:
							self.proposals.append(VariantProposal(variant))
					self.hint = None
					return
				
			elif state == self.S_COMMAND:
				if token.type == Token.T_END:
					# \beg
					#     ^
					for command in self.__command_store.find(self.subprefix):
						for variant in command.variants:
							self.proposals.append(VariantProposal(variant))
					self.hint = None
					
					return
				else:
					# find the command exactly matching the prefix
					try:
						command = self.__command_store[self.subprefix]
					except KeyError:
						debug(self, "parse", "Command not found: %s" % self.subprefix)
						return
					
					if token.type == Token.T_CURLY_OPEN:
						variant_signature.append(MandatoryArgument)
						state = self.S_MANDATORY_ARGUMENT_START
					elif token.type == Token.T_SQUARE_OPEN:
						variant_signature.append(OptionalArgument)
						state = self.S_OPTIONAL_ARGUMENT_START
					elif token.type == Token.T_ROUND_OPEN:
						variant_signature.append(VectorArgument)
						state = self.S_VECTOR_ARGUMENT_START
						component_counter = 0
					else:
						# unexpected token
						return
			
			elif state == self.S_MANDATORY_ARGUMENT_START:
				
				if token.type == Token.T_END:
					# \begin{
					#        ^
					self.subprefix = ""
					try:
						variant = command.get_variant_by_signature(variant_signature)
						argument = variant.get_used_argument(len(variant_signature) - 1)
						
						self.proposals = [ChoiceProposal(c) for c in argument.choices]
						self.hint = argument.label
						
						return
							
					except KeyError:
						debug(self, "parse", "Command variant for %s not found" % command.name)
						return
				elif token.type == Token.T_CURLY_CLOSE:
					state = self.S_COMMAND
				elif token.type == Token.T_NAME:
					self.subprefix = token.value
					state = self.S_MANDATORY_ARGUMENT
			
			elif state == self.S_MANDATORY_ARGUMENT:
				if token.type == Token.T_END:
					# \begin{ab
					#          ^
					
					try:
						variant = command.get_variant_by_signature(variant_signature)
						argument = variant.get_used_argument(len(variant_signature) - 1)
						
						self.proposals = [ChoiceProposal(c) for c in argument.choices if c.value.startswith(self.subprefix)]
						self.hint = argument.label
						
						return
							
					except KeyError:
						debug(self, "parse", "Command variant for %s not found" % command.name)
						return
				elif token.type == Token.T_CURLY_CLOSE:
					state = self.S_COMMAND
			
			elif state == self.S_OPTIONAL_ARGUMENT_START:
				if token.type == Token.T_END:
					# \bibitem[
					#          ^
					try:
						variant = command.get_variant_by_signature(variant_signature)
						argument = variant.get_used_argument(len(variant_signature) - 1)
						
						self.proposals = [ChoiceProposal(c) for c in argument.choices]
						self.hint = argument.label
						
						return
							
					except KeyError:
						debug(self, "parse", "Command variant for %s not found" % command.name)
						return
				elif token.type == Token.T_SQUARE_CLOSE:
					state = self.S_COMMAND
				elif token.type == Token.T_NAME:
					state = self.S_OPTIONAL_ARGUMENT
			
			elif state == self.S_OPTIONAL_ARGUMENT:
				if token.type == Token.T_END:
					# \bibitem[ab
					#            ^
					try:
						variant = command.get_variant_by_signature(variant_signature)
						argument = variant.get_used_argument(len(variant_signature) - 1)
						
						self.proposals = [ChoiceProposal(c) for c in argument.choices if c.value.startswith(self.subprefix)]
						self.hint = argument.label
						
						return
							
					except KeyError:
						log(self, "parse", "Command variant for %s not found" % command.name)
						return
				elif token.type == Token.T_SQUARE_CLOSE:
					state = self.S_COMMAND
			
			elif state == self.S_VECTOR_ARGUMENT_START:
				if token.type == Token.T_END:
					# \makebox(
					#          ^
					debug(self, "parse", "Searching for variant " + str(variant_signature))
					try:
						variant = command.get_variant_by_signature(variant_signature)
						
						debug(self, "parse", "Found " + str(variant))
						
						argument = variant.get_used_argument(len(variant_signature) - 1)
						
						self.hint = argument.components[0].label
						
						return
							
					except KeyError:
						debug(self, "parse", "Command variant for %s not found" % command.name)
						return
				elif token.type == Token.T_ROUND_CLOSE:
					state = self.S_COMMAND
				elif token.type == Token.T_NAME or token.type == Token.T_OTHER:
					state = self.S_VECTOR_ARGUMENT
			
			elif state == self.S_VECTOR_ARGUMENT:
				if token.type == Token.T_END:
					# \makebox(2cm,
					#              ^
					try:
						variant = command.get_variant_by_signature(variant_signature)
						argument = variant.get_used_argument(len(variant_signature) - 1)
						
						self.hint = argument.components[component_counter].label
						
						return
							
					except KeyError:
						debug(self, "parse", "Command variant for %s not found" % command.name)
						return
				elif token.type == Token.T_COMMA:
					component_counter += 1
				elif token.type == Token.T_ROUND_CLOSE:
					state = self.S_COMMAND
			
			else:
				debug(self, "parse", "Unexpected token")
				return
			