# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2007 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

from gobject import timeout_add, source_remove, child_watch_add, SIGNAL_RUN_LAST, TYPE_NONE, \
		TYPE_STRING, GObject, type_register, TYPE_BOOLEAN, TYPE_INT, TYPE_PYOBJECT
from subprocess import Popen, PIPE, STDOUT
from os import kill, system, chdir, O_NONBLOCK
from os.path import splitext, dirname
from signal import SIGTERM
from fcntl import fcntl, F_SETFL

from LaTeXPlugin.Installation import Installation
from LaTeXPlugin.Logger import debug, error
from Processor import Processor
from RubberProcessor import RubberProcessor
from LatexProcessor import LatexProcessor
from ChktexProcessor import ChktexProcessor
from Message import *


class BuildEngine(GObject):
	"""
	"""
	
	OUTPUT_POLL_INTERVAL = 250
	T_STDOUT, T_COMMAND = 0, 1
	
	__gsignals__ = {
		"output" : (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_STRING, TYPE_INT])),		# output, T_STDOUT || T_COMMAND
		"job-started" : (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_INT])),
		"job-finished" : (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_INT, TYPE_STRING, TYPE_PYOBJECT])),	# job id, summary, messages
		"job-failed" : (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_INT, TYPE_STRING, TYPE_PYOBJECT])),		# job id, summary, messages
		"job-aborted" : (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_INT]))		# job id
	}
	
	def __init__(self):
		GObject.__init__(self)
		
		self.__process = None
		self.__processors = {}
	
	def set_profile(self, profile):
		self.__profile = profile
		self.__job_iter = profile.jobs.__iter__()
	
	def set_filename(self, filename):
		self.__filename = filename
		self.__filename_stub = splitext(filename)[0]
		self.__replacements = { r"%f" : filename, r"%e": self.__filename_stub, r"%p" : Installation().path }
	
	def run(self):
		chdir(dirname(self.__filename))
		
		self.__job_counter = 0
		self.__proceed()
	
	def abort(self):
		if self.__process:
			source_remove(self.__id_output)
			source_remove(self.__id_exit)
			
			try:
				kill(self.__process.pid, SIGTERM)
			except OSError:
				# no such pid
				pass
			
			self.emit("job-aborted", self.__job_counter)
	
	def __proceed(self):
		try:
			self.__job = self.__job_iter.next()
			
			command = self.__job.get_command(self.__replacements)
			
			self.emit("job-started", self.__job_counter)
			
			self.emit("output", command, self.T_COMMAND)
			
			# run child process
			self.__process = Popen(command, shell=True, stdout=PIPE, stderr=STDOUT)
			
			# unblock pipes
			fcntl(self.__process.stdout, F_SETFL, O_NONBLOCK)
			
			# monitor process and pipes
			self.__id_output = timeout_add(self.OUTPUT_POLL_INTERVAL, self.__read_process_output)
			self.__id_exit = child_watch_add(self.__process.pid, self.__process_exited)
			
		except StopIteration:
			#
			# no more jobs, view output
			#
			if len(self.__profile.viewer):
				self.__open_result()
		
	def __process_exited(self, pid, condition):
		source_remove(self.__id_exit)
		source_remove(self.__id_output)
		
		# read remaining output
		try:
			s = self.__process.stdout.read()
			self.emit("output", s, self.T_STDOUT)
		except IOError:
			pass
		
		# get processor
		processor = self.__get_processor(self.__job.processor)
		
		# use processor
		processor.filename = self.__filename
		processor.condition = condition
		
		summary = processor.get_summary()
		messages = processor.get_messages()
		
		
		if condition:
			self.emit("job-failed", self.__job_counter, summary, messages)
			if self.__job.stopper:
				# abort build process and view output
				self.__open_result()
				return
		else:
			self.emit("job-finished", self.__job_counter, summary, messages)
		
		self.__job_counter += 1
		self.__proceed()
	
	def __get_processor(self, id):
		"""
		Return a processor for a given id (e.g. LatexProcessor for "latex")
		"""
		try:
			# try to select instance
			processor = self.__processors[id]
		except KeyError:
			# create new processor
			if id == "rubber":
				processor = RubberProcessor()
			elif id == "latex":
				processor = LatexProcessor()
			elif id == "chktex":
				processor = ChktexProcessor()
			else:
				processor = Processor()
			
			self.__processors[id] = processor
		
		return processor
	
	def __read_process_output(self):
		try:
			s = self.__process.stdout.read()
			self.emit("output", s, self.T_STDOUT)
		except IOError:
			pass
		
		return True
	
	def __open_result(self):
		# TODO: we need some checking here but getting mtime is hard because we
		# don't know the resulting filename
		viewer_command = self.__profile.viewer.replace(r"%e", self.__filename_stub)
		debug(self, "__open_result", viewer_command)
		system(viewer_command)

type_register(BuildEngine)