# -*- coding: utf-8 -*-

# This file is part of the Gedit LaTeX Plugin
#
# Copyright (C) 2007 Michael Zeising
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public Licence as published by the Free Software
# Foundation; either version 2 of the Licence, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence for more 
# details.
#
# You should have received a copy of the GNU General Public Licence along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA  02110-1301, USA

from gnomevfs import get_mime_type, get_local_path_from_uri
from gobject import GObject, type_register, SIGNAL_RUN_LAST, TYPE_NONE, TYPE_INT

from LaTeXPlugin.Logger import debug
from LatexLanguageHelper import LatexLanguageHelper
from BibtexLanguageHelper import BibtexLanguageHelper
from MetaFile import MetaFile


LANGUAGE_NONE, LANGUAGE_LATEX, LANGUAGE_BIBTEX = 0, 1, 2

_language = ["none", "latex", "bibtex"]


class DocumentHelper(GObject):
	"""
	This augments the functionality of a GeditView
	"""
	
	__gsignals__ = {
		"language-changed" : (SIGNAL_RUN_LAST, TYPE_NONE, ([TYPE_INT])),
		"source-changed" : (SIGNAL_RUN_LAST, TYPE_NONE, ([])),
		"document-changed" : (SIGNAL_RUN_LAST, TYPE_NONE, ([]))		# file opened, saved
	}
	
	def __init__(self, document_helper_manager, gedit_view):
		debug(self, "__init__")
		
		GObject.__init__(self)
		
		self.__document_helper_manager = document_helper_manager
		self.__view = gedit_view
		self.__buffer = self.__view.get_buffer()
		
		self.__meta_file = None
		
		self.__changed = True
		self.__line = 0
		
		self.__language = LANGUAGE_NONE
		self.__check_language = True
		self.__language_helper = None
		
		self.__do_check_language()
		
		self.__buffer.connect("loaded", self.__buffer_loaded)
		self.__buffer.connect("saved", self.__buffer_saved)
		self.__buffer.connect("changed", self.__buffer_changed)
		
		self.__view.connect("key-release-event", self.__key_released)
	
	def __key_released(self, view, event):
		line = self.__buffer.get_iter_at_mark(self.__buffer.get_insert()).get_line()
		if line == self.__line:
			return
		
		self.__line = line
		
		if not self.__changed:
			return
		
		self.__changed = False
		
		self.emit("source-changed")
	
	def __buffer_changed(self, buffer):
		self.__changed = True
	
	def __buffer_loaded(self, *args):
		debug(self, "__buffer_loaded")
		
		if self.__check_language:
			self.__do_check_language()
			
		self.emit("document-changed")
	
	def __buffer_saved(self, *args):
		if self.__check_language:
			self.__do_check_language()
			
		self.emit("document-changed")
	
	def __do_check_language(self):
		language = self.__guess_language()
		
		debug(self, "__do_check_language", 
				"Guessed language {%s}" % _language[language])
		
		if language != self.__language:
			debug(self, "__do_check_language", 
				"Language changed from {%s} to {%s}" % (_language[self.__language], 
													_language[language]))
			
			self.__set_language(language)
			
			self.emit("language-changed", language)
	
	def __set_language(self, language):
		if self.__language == LANGUAGE_NONE:
			if language == LANGUAGE_LATEX:
				self.__language_helper = LatexLanguageHelper(self)
				self.__set_highlighting(LANGUAGE_LATEX)
			elif language == LANGUAGE_BIBTEX:
				self.__language_helper = BibtexLanguageHelper(self)
		
		self.__language = language
	
	def __guess_language(self):
		"""
		Guess language from contained MIME type
		"""
		try:
			uri = self.__buffer.get_uri()
			if not uri:
				return LANGUAGE_NONE
			
			mime_type = get_mime_type(uri)
			
			if mime_type == "text/x-tex":
				return LANGUAGE_LATEX
			elif mime_type == "text/x-bibtex":
				return LANGUAGE_BIBTEX
			else:
				return LANGUAGE_NONE
		except RuntimeError:
			return LANGUAGE_NONE
	
	def __set_highlighting(self, language):
		if language == LANGUAGE_LATEX:
			try:
				from gtksourceview import SourceLanguagesManager
				
				language = SourceLanguagesManager().get_language_from_mime_type("text/x-tex")
				self.__buffer.set_language(language)
				self.__buffer.set_highlight(True)
			except ImportError:
				# import of gtksourceview failed, gtksourceview2 seems to be installed...
				try:
					from gtksourceview2 import LanguageManager
					
					manager = LanguageManager()
					
					language = None
					
					# the API is not frozen yet...
					try:
						language = manager.get_language_by_id("latex")
					except AttributeError:
						language = manager.get_language("latex")
						
					if language:
						self.__buffer.set_language(language)
						try:
							self.__buffer.set_highlight(True)
						except AttributeError:
							self.__buffer.set_highlight_syntax(True)
						
				except ImportError, s:
					error(self, "enable_highlighting", s)
	
	def set_language(self, language):
		self.__check_language = False
		
		if language != self.__language:
			self.__set_language(language)
			
			self.emit("language-changed", language)
	
	def get_language(self):
		return self.__language
	
	@property
	def language_helper(self):
		return self.__language_helper
	
	@property
	def filename(self):
		uri = self.__buffer.get_uri()
		if uri:
			try:
				return get_local_path_from_uri(uri)
			except RuntimeError:
				return None
		return None
	
	def select_range(self, start, end):
		it_start = self.__buffer.get_iter_at_offset(start)
		it_end = self.__buffer.get_iter_at_offset(end)
		self.__buffer.select_range(it_start, it_end)
		self.__view.scroll_to_iter(it_start, .25)
		self.__view.grab_focus()
	
	@property
	def view(self):
		"""
		Return the gedit.View wrapped by this Helper
		"""
		return self.__view
	
	@property
	def document_helper_manager(self):
		return self.__document_helper_manager
	
	@property
	def text(self):
		charset = self.__buffer.get_encoding().get_charset()
		debug(self, "__get_text", "character set is '%s'" % charset)
		return self.__buffer.get_text(self.__buffer.get_start_iter(), \
										 self.__buffer.get_end_iter(), False).decode(charset)
	
	@property
	def cursor_offset(self):
		return self.__buffer.get_iter_at_mark(self.__buffer.get_insert()).get_offset()
	
	def jump_to_line(self, line):
		debug(self, "jump_to_line", line)
		
		it = self.__buffer.get_iter_at_line(line)
		self.__view.scroll_to_iter(it, .25)
		self.__buffer.goto_line(line)
		self.__view.grab_focus()
	
	@property
	def meta_file(self):
		filename = self.filename
		if not filename:
			debug(self, "meta_file", "No filename")
			return None
		
		if not self.__meta_file:
			self.__meta_file = MetaFile(filename)
		
		return self.__meta_file
	
	
type_register(DocumentHelper)
